# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse
import sys

from colcon_core.package_selection import add_arguments \
    as add_packages_arguments
from colcon_core.package_selection import get_package_descriptors
from colcon_core.package_selection import select_package_decorators
from colcon_core.plugin_system import satisfies_version
from colcon_core.topological_order import topological_order_packages
from colcon_core.verb import VerbExtensionPoint


class InfoVerb(VerbExtensionPoint):
    """Package information."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(VerbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            'package_names', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Only show the information of a subset of packages')

        # only added so that package selection arguments can be used
        # which use the build directory to store state information
        parser.add_argument(
            '--build-base',
            default='build',
            help='The base path for all build directories (default: build)')

        add_packages_arguments(parser)

    def main(self, *, context):  # noqa: D102
        descriptors = get_package_descriptors(
            context.args, additional_argument_names=['*'])
        decorators = topological_order_packages(
            descriptors, recursive_categories=('run', ))
        select_package_decorators(context.args, decorators)

        if context.args.package_names:
            all_package_names = {d.descriptor.name for d in decorators}
            # warn about passed package names which are unknown
            for pkg_name in context.args.package_names:
                if pkg_name not in all_package_names:
                    print(
                        "Package '{pkg_name}' not found".format_map(locals()),
                        file=sys.stderr)
            # filter decorators using passed package names
            decorators = [
                d for d in decorators
                if d.descriptor.name in context.args.package_names]
            if not decorators:
                return 1
        if not decorators:
            return 'No packages found'

        for decorator in sorted(decorators, key=lambda d: d.descriptor.name):
            if not decorator.selected:
                continue
            pkg = decorator.descriptor
            print('path:', pkg.path)
            print('  type:', pkg.type)
            print('  name:', pkg.name)
            if pkg.dependencies:
                print('  dependencies:')
                for category in sorted(pkg.dependencies.keys()):
                    print(
                        '    {category}:'.format_map(locals()),
                        ' '.join(sorted(pkg.dependencies[category])))
            if pkg.hooks:
                print('  hooks:', ' '.join(pkg.hooks))
            if pkg.metadata:
                print('  metadata:')
                for key in sorted(pkg.metadata.keys()):
                    value = pkg.metadata[key]
                    print(
                        '    {key}: {value}'
                        .format_map(locals()))


def argument_package_name(value):
    """
    Check if an argument is a valid package name.

    Used as a ``type`` callback in ``add_argument()`` calls.
    Package names starting with a dash must be prefixed with a space to avoid
    collisions with command line arguments.

    :param str value: The command line argument
    :returns: The package name
    :raises argparse.ArgumentTypeError: if the value starts with a dash
    """
    if value.startswith('-'):
        raise argparse.ArgumentTypeError('unrecognized argument: ' + value)
    return value.lstrip()



# Copyright 2024 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from unittest.mock import patch

from colcon_core.dependency_descriptor import DependencyDescriptor
from colcon_core.package_descriptor import PackageDescriptor
from colcon_package_information.package_augmentation.\
    check_dependency_constraint \
    import CheckDependencyConstraintPackageAugmentation
import pytest


@pytest.mark.parametrize(
    'operator,expect_lt,expect_eq,expect_gt', [
        (None,          0, 0, 0),
        ('version_eq',  1, 0, 1),
        ('version_gt',  1, 1, 0),
        ('version_gte', 1, 0, 0),
        ('version_lt',  0, 1, 1),
        ('version_lte', 0, 0, 1),
        ('version_neq', 0, 1, 0),
    ])
def test_operators(operator, expect_lt, expect_eq, expect_gt):
    pkg_a_v1 = PackageDescriptor('/tmp/pkg_a')
    pkg_a_v1.name = 'pkg_a'
    pkg_a_v1.metadata['version'] = '1.0'

    pkg_a_v2 = PackageDescriptor('/tmp/pkg_a')
    pkg_a_v2.name = 'pkg_a'
    pkg_a_v2.metadata['version'] = '2.0'

    pkg_a_v3 = PackageDescriptor('/tmp/pkg_a')
    pkg_a_v3.name = 'pkg_a'
    pkg_a_v3.metadata['version'] = '3.0'

    pkg_a_dep = DependencyDescriptor('pkg_a')
    if operator is not None:
        pkg_a_dep.metadata[operator] = '2.0'

    pkg_b = PackageDescriptor('/tmp/pkg_b')
    pkg_b.name = 'pkg_b'
    pkg_b.dependencies['build'] = {pkg_a_dep}

    extension = CheckDependencyConstraintPackageAugmentation()

    with patch(
        'colcon_package_information.package_augmentation.'
        'check_dependency_constraint.logger.warning'
    ) as warning:
        extension.augment_packages({pkg_a_v1, pkg_b})
        assert expect_lt == warning.call_count
        warning.reset_mock()

        extension.augment_packages({pkg_a_v2, pkg_b})
        assert expect_eq == warning.call_count
        warning.reset_mock()

        extension.augment_packages({pkg_a_v3, pkg_b})
        assert expect_gt == warning.call_count
        warning.reset_mock()


def test_dependency_missing():
    pkg_b = PackageDescriptor('/tmp/pkg_b')
    pkg_b.name = 'pkg_b'
    pkg_b.dependencies['build'] = {
        DependencyDescriptor('pkg_a'),
    }

    extension = CheckDependencyConstraintPackageAugmentation()
    extension.augment_packages({pkg_b})


def test_dependency_missing_version():
    pkg_a = PackageDescriptor('/tmp/pkg_a')
    pkg_a.name = 'pkg_a'

    pkg_b = PackageDescriptor('/tmp/pkg_b')
    pkg_b.name = 'pkg_b'
    pkg_b.dependencies['build'] = {
        DependencyDescriptor('pkg_a'),
    }

    extension = CheckDependencyConstraintPackageAugmentation()
    extension.augment_packages({pkg_a, pkg_b})


def test_dependency_invalid_version():
    pkg_a = PackageDescriptor('/tmp/pkg_a')
    pkg_a.name = 'pkg_a'
    pkg_a.metadata['version'] = 'totally!invalid&version'

    pkg_b = PackageDescriptor('/tmp/pkg_b')
    pkg_b.name = 'pkg_b'
    pkg_b.dependencies['build'] = {
        DependencyDescriptor('pkg_a'),
    }

    extension = CheckDependencyConstraintPackageAugmentation()
    extension.augment_packages({pkg_a, pkg_b})


def test_dependency_invalid_version_operator():
    pkg_a = PackageDescriptor('/tmp/pkg_a')
    pkg_a.name = 'pkg_a'
    pkg_a.metadata['version'] = '2.0'

    pkg_b = PackageDescriptor('/tmp/pkg_b')
    pkg_b.name = 'pkg_b'
    pkg_b.dependencies['build'] = {
        DependencyDescriptor('pkg_a', metadata={
            'version_eq': 'totally!invalid&version'
        }),
    }

    extension = CheckDependencyConstraintPackageAugmentation()
    extension.augment_packages({pkg_a, pkg_b})


@pytest.mark.parametrize(
    'a_versions,skip_incompatible,dep_categories,expect_present', [
        (['1.0'],        True,  {'build'},        False),
        (['1.0', '3.0'], True,  {'build'},        False),
        (['1.0', '2.0'], True,  {'build'},        True),
        (['1.0'],        False, {'build'},        True),
        (['1.0'],        True,  {'build', 'run'}, False),
        ([None],         True,  {'build'},        True),
    ])
def test_skip_incompatible(
    a_versions, skip_incompatible, dep_categories, expect_present
):
    descs = set()
    for i, v in enumerate(a_versions):
        pkg_a = PackageDescriptor(f'/tmp/pkg_a_{i}')
        pkg_a.name = 'pkg_a'
        if v is not None:
            pkg_a.metadata['version'] = v
        descs.add(pkg_a)

    metadata = {
        'version_eq': '2.0',
        'skip_incompatible': skip_incompatible,
        'categories': dep_categories,
    }
    pkg_a_dep = DependencyDescriptor('pkg_a', metadata=metadata)

    pkg_b = PackageDescriptor('/tmp/pkg_b')
    pkg_b.name = 'pkg_b'
    for cat in dep_categories:
        pkg_b.dependencies[cat] = {pkg_a_dep}

    descs.add(pkg_b)

    extension = CheckDependencyConstraintPackageAugmentation()
    extension.augment_packages(descs)

    for cat in dep_categories:
        if expect_present:
            assert pkg_a_dep in pkg_b.dependencies[cat]
        else:
            assert pkg_a_dep not in pkg_b.dependencies.get(cat, set())



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import operator

from colcon_core.package_augmentation import logger as \
    package_augmentation_logger
from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.plugin_system import satisfies_version
from packaging.version import Version

logger = package_augmentation_logger.getChild('check_dependency_constraint')


class CheckDependencyConstraintPackageAugmentation(
    PackageAugmentationExtensionPoint
):
    """Check package dependency constraints."""

    # the priority needs to be lower than other extensions to allow them to
    # augment dependency information before
    PRIORITY = 1

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def augment_packages(  # noqa: D102
        self, descs, *, additional_argument_names=None
    ):
        descs_dict = {}
        for desc in descs:
            descs_dict.setdefault(desc.name, []).append(desc)
        for desc in descs:
            deps = desc.get_dependencies()
            for dep in deps:
                # skip dependencies which don't have a descriptor
                if dep.name not in descs_dict:
                    continue

                miss = all(
                    self._check_version_constraints(desc, dep, dep_desc)
                    for dep_desc in descs_dict.get(dep.name, ()))
                if miss and dep.metadata.get('skip_incompatible'):
                    logger.warning(
                        'Dropping version-incompatible dependency on ' +
                        dep.name + ' from ' + desc.name)
                    for category in dep.metadata.get('categories') or ():
                        desc.dependencies[category].discard(dep)

    def _check_version_constraints(self, desc, dep, dep_desc):
        """
        Check dependency version constraints against another package.

        :param desc: The package declaring the dependency being checked
        :param dep: The dependency being checked
        :param dep_desc: The package to which the dependency refers

        :returns: True if one or more constraints fail, False if all
          constraints pass, and None if no constraints were checked.
        """
        # if the dependency descriptor doesn't have a version there is nothing
        # to compare to
        if 'version' not in dep_desc.metadata:
            return
        try:
            dep_version = Version(dep_desc.metadata['version'])
        except Exception:  # noqa: B902
            # skip check if the version fails to parse
            return

        def lte(a, b):
            return operator.lt(a, b) or operator.eq(a, b)

        def gte(a, b):
            return operator.gt(a, b) or operator.eq(a, b)

        operators = {
            'version_lt': (operator.lt, 'less than'),
            'version_lte': (lte, 'less than or equal to'),
            'version_eq': (operator.eq, 'equal to'),
            'version_neq': (operator.ne, 'not equal to'),
            'version_gte': (gte, 'greater than or equal to'),
            'version_gt': (operator.gt, 'greater than'),
        }
        miss = None
        for key, value in dep.metadata.items():
            # only consider version operator metadata
            if key not in operators:
                continue
            op, msg = operators[key]
            try:
                version_constraint = Version(value)
            except Exception:  # noqa: B902
                logger.error(
                    "Failed to parse version '" + value + "' with " +
                    "constraint '" + msg + "' for dependency " + dep.name +
                    ' in package ' + desc.name)
                # skip check if the version fails to parse
                continue

            if not op(dep_version, version_constraint):
                miss = True
                logger.warning(
                    desc.name + ' depends on ' + dep.name +
                    ' which has version ' + dep_desc.metadata['version'] +
                    ' but expects it to be ' + msg + ' ' + value)
            else:
                miss = miss or False
                logger.debug(
                    desc.name + ' depends on ' + dep.name +
                    ' which has version ' + dep_desc.metadata['version'] +
                    ' which satisfies to be ' + msg + ' ' + value)
        return miss



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import defaultdict
from collections import OrderedDict
import itertools
import os
from pathlib import Path

from colcon_core.package_selection import add_arguments \
    as add_packages_arguments
from colcon_core.package_selection import get_package_descriptors
from colcon_core.package_selection import select_package_decorators
from colcon_core.plugin_system import satisfies_version
from colcon_core.topological_order import topological_order_packages
from colcon_core.verb import VerbExtensionPoint


class GraphVerb(VerbExtensionPoint):
    """Generate a visual representation of the dependency graph."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(VerbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        # only added so that package selection arguments can be used
        # which use the build directory to store state information
        parser.add_argument(
            '--build-base',
            default='build',
            help='The base path for all build directories (default: build)')

        add_packages_arguments(parser)

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--dot',
            action='store_true',
            default=False,
            help='Output topological graph in DOT '
                 '(e.g. pass the output to dot: ` | dot -Tpng -o graph.png`), '
                 'legend: blue=build, red=run, tan=test, dashed=indirect')

        group.add_argument(
            '--density',
            action='store_true',
            default=False,
            help='Output density of the graph (only without --dot)')

        parser.add_argument(
            '--legend',
            action='store_true',
            default=False,
            help='Output legend for the graph')

        parser.add_argument(
            '--dot-cluster',
            action='store_true',
            default=False,
            help='Cluster packages by their filesystem path (only affects '
                 '--dot)')
        parser.add_argument(
            '--dot-include-skipped',
            action='store_true',
            default=False,
            help='Also output skipped packages (only affects --dot)')

    def main(self, *, context):  # noqa: D102
        args = context.args

        descriptors = get_package_descriptors(args)

        decorators = topological_order_packages(
            descriptors, recursive_categories=('run', ))

        select_package_decorators(args, decorators)

        if not args.dot:
            if args.legend:
                print('+ marks when the package in this row can be processed')
                print('* marks a direct dependency '
                      'from the package indicated by the + in the same column '
                      'to the package in this row')
                print('. marks a transitive dependency')
                print()

            # draw dependency graph in ASCII
            shown_decorators = list(filter(lambda d: d.selected, decorators))
            max_length = max([
                len(m.descriptor.name) for m in shown_decorators] + [0])
            lines = [
                m.descriptor.name.ljust(max_length + 2)
                for m in shown_decorators]
            depends = [
                m.descriptor.get_dependencies() for m in shown_decorators]
            rec_depends = [
                m.descriptor.get_recursive_dependencies(
                    [d.descriptor for d in decorators],
                    recursive_categories=('run', ))
                for m in shown_decorators]

            empty_cells = 0
            for i, decorator in enumerate(shown_decorators):
                for j in range(len(lines)):
                    if j == i:
                        # package i is being processed
                        lines[j] += '+'
                    elif shown_decorators[j].descriptor.name in depends[i]:
                        # package i directly depends on package j
                        lines[j] += '*'
                    elif shown_decorators[j].descriptor.name in rec_depends[i]:
                        # package i recursively depends on package j
                        lines[j] += '.'
                    else:
                        # package i doesn't depend on package j
                        lines[j] += ' '
                        empty_cells += 1
            if args.density:
                empty_fraction = \
                    empty_cells / (len(lines) * (len(lines) - 1)) \
                    if len(lines) > 1 else 1.0
                # normalize to 200% since half of the matrix should be empty
                density_percentage = 200.0 * (1.0 - empty_fraction)
                print('dependency density %.2f %%' % density_percentage)
                print()

        else:  # --dot
            lines = ['digraph graphname {']

            decorators_by_name = defaultdict(set)
            for deco in decorators:
                decorators_by_name[deco.descriptor.name].add(deco)

            selected_pkg_names = [
                m.descriptor.name for m in decorators
                if m.selected or args.dot_include_skipped]
            has_duplicate_names = \
                len(selected_pkg_names) != len(set(selected_pkg_names))
            selected_pkg_names = set(selected_pkg_names)

            # collect selected package decorators and their parent path
            nodes = OrderedDict()
            for deco in reversed(decorators):
                if deco.selected or args.dot_include_skipped:
                    nodes[deco] = Path(deco.descriptor.path).parent

            # collect direct dependencies
            direct_edges = defaultdict(set)
            for deco in reversed(decorators):
                if (
                    not deco.selected and
                    not args.dot_include_skipped
                ):
                    continue
                # iterate over dependency categories
                for category, deps in deco.descriptor.dependencies.items():
                    # iterate over dependencies
                    for dep in deps:
                        if dep not in selected_pkg_names:
                            continue
                        # store the category of each dependency
                        # use the decorator
                        # since there might be packages with the same name
                        direct_edges[(deco, dep)].add(category)

            # collect indirect dependencies
            indirect_edges = defaultdict(set)
            for deco in reversed(decorators):
                if not deco.selected:
                    continue
                # iterate over dependency categories
                for category, deps in deco.descriptor.dependencies.items():
                    # iterate over dependencies
                    for dep in deps:
                        # ignore direct dependencies
                        if dep in selected_pkg_names:
                            continue
                        # ignore unknown dependencies
                        if dep not in decorators_by_name.keys():
                            continue
                        # iterate over recursive dependencies
                        for rdep in itertools.chain.from_iterable(
                            d.recursive_dependencies
                            for d in decorators_by_name[dep]
                        ):
                            if rdep not in selected_pkg_names:
                                continue
                            # skip edges which are redundant to direct edges
                            if (deco, rdep) in direct_edges:
                                continue
                            indirect_edges[(deco, rdep)].add(category)

            try:
                # HACK Python 3.5 can't handle Path objects
                common_path = os.path.commonpath(
                    [str(p) for p in nodes.values()])
            except ValueError:
                common_path = None

            def get_node_data(decorator):
                if not has_duplicate_names:
                    # use name where possible so the dot code is easy to read
                    return decorator.descriptor.name, \
                        '' if (
                            decorator.selected or
                            not args.dot_include_skipped
                        ) else '[color = "gray" fontcolor = "gray"]'
                # otherwise append the descriptor id to make each node unique
                descriptor_id = id(decorator.descriptor)
                return (
                    '{decorator.descriptor.name}_{descriptor_id}'
                    .format_map(locals()),
                    ' [label = "{decorator.descriptor.name}"]'
                    .format_map(locals()),
                )

            if not args.dot_cluster or common_path is None:
                # output nodes
                for deco in nodes.keys():
                    if (
                        not deco.selected and
                        not args.dot_include_skipped
                    ):
                        continue
                    node_name, attributes = get_node_data(deco)
                    lines.append(
                        '  "{node_name}"{attributes};'.format_map(locals()))
            else:
                # output clusters
                clusters = defaultdict(set)
                for deco, path in nodes.items():
                    clusters[path.relative_to(common_path)].add(deco)
                for i, cluster in zip(range(len(clusters)), clusters.items()):
                    path, decos = cluster
                    if path.name:
                        # wrap cluster in subgraph
                        lines.append(
                            '  subgraph cluster_{i} {{'.format_map(locals()))
                        lines.append(
                            '    label = "{path}";'.format_map(locals()))
                        indent = '    '
                    else:
                        indent = '  '
                    for deco in decos:
                        node_name, attributes = get_node_data(deco)
                        lines.append(
                            '{indent}"{node_name}"{attributes};'
                            .format_map(locals()))
                    if path.name:
                        lines.append('  }')

            # output edges
            color_mapping = OrderedDict((
                ('build', '#0000ff'),  # blue
                ('run', '#ff0000'),  # red
                ('test', '#d2b48c'),  # tan
            ))
            for style, edges in zip(
                ('', ', style="dashed"'),
                (direct_edges, indirect_edges),
            ):
                for (deco_start, node_end), categories in edges.items():
                    start_name, _ = get_node_data(deco_start)
                    for deco in decorators_by_name[node_end]:
                        end_name, _ = get_node_data(deco)
                        edge_alpha = '' \
                            if deco_start.selected and deco.selected else '77'
                        colors = ':'.join([
                            color + edge_alpha
                            for category, color in color_mapping.items()
                            if category in categories])
                        lines.append(
                            '  "{start_name}" -> "{end_name}" '
                            '[color="{colors}"{style}];'.format_map(locals()))

            if args.legend:
                lines.append('  subgraph cluster_legend {')
                lines.append('    color=gray')
                lines.append('    label="Legend";')
                lines.append('    margin=0;')
                # invisible nodes between the dependency edges
                lines.append('    node [label="", shape=none];')

                previous_node = '_legend_first'
                # an edge for each dependency type
                for dependency_type, color in color_mapping.items():
                    next_node = '_legend_' + dependency_type
                    lines.append(
                        '    {previous_node} -> {next_node} '
                        '[label="{dependency_type} dep.", color="{color}"];'
                        .format_map(locals()))
                    previous_node = next_node
                lines.append(
                    '    {previous_node} -> _legend_last '
                    '[label="indirect dep.", style="dashed"];'
                    .format_map(locals()))

                # layout all legend nodes on the same rank
                lines.append('    {')
                lines.append('      rank=same;')
                lines.append('      _legend_first;')
                for dependency_type in color_mapping.keys():
                    lines.append(
                        '      _legend_{dependency_type};'
                        .format_map(locals()))
                lines.append('      _legend_last;')
                lines.append('    }')

                lines.append('  }')

            lines.append('}')

        for line in lines:
            print(line)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.command import get_prog_name
from colcon_core.package_selection import add_arguments \
    as add_packages_arguments
from colcon_core.package_selection import get_package_descriptors
from colcon_core.package_selection import select_package_decorators
from colcon_core.plugin_system import satisfies_version
from colcon_core.topological_order import topological_order_packages
from colcon_core.verb import VerbExtensionPoint


class ListVerb(VerbExtensionPoint):
    """List packages, optionally in topological ordering."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(VerbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        # only added so that package selection arguments can be used
        # which use the build directory to store state information
        parser.add_argument(
            '--build-base',
            default='build',
            help='The base path for all build directories (default: build)')

        add_packages_arguments(parser)

        parser.add_argument(
            '--topological-order', '-t',
            action='store_true',
            default=False,
            help='Order output based on topological ordering (breadth-first)')

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--names-only', '-n',
            action='store_true',
            default=False,
            help='Output only the name of each package but not the path')
        group.add_argument(
            '--paths-only', '-p',
            action='store_true',
            default=False,
            help='Output only the path of each package but not the name')

        group = parser.add_argument_group('Obsolete arguments')
        command_name = get_prog_name()
        self._add_obsolete_argument(
            group,
            '--topological-graph', '-g',
            action='store_true',
            default=False,
            help='Instead use `{command_name} graph`'.format_map(locals()))
        self._add_obsolete_argument(
            group,
            '--topological-graph-dot',
            action='store_true',
            default=False,
            help='Instead use `{command_name} graph --dot`'
                 .format_map(locals()))
        self._add_obsolete_argument(
            group,
            '--topological-graph-density',
            action='store_true',
            default=False,
            help='Instead use `{command_name} graph --density`'
                 .format_map(locals()))
        self._add_obsolete_argument(
            group,
            '--topological-graph-legend',
            action='store_true',
            default=False,
            help='Instead use `{command_name} graph --legend`'
                 .format_map(locals()))
        self._add_obsolete_argument(
            group,
            '--topological-graph-dot-cluster',
            action='store_true',
            default=False,
            help='Instead use `{command_name} graph --dot --dot-cluster`'
                 .format_map(locals()))
        self._add_obsolete_argument(
            group,
            '--topological-graph-dot-include-skipped',
            action='store_true',
            default=False,
            help='Instead use `{command_name} graph --dot '
                 '--dot-include-skipped`'.format_map(locals()))

    def _add_obsolete_argument(self, group, *args, **kwargs):
        arg = group.add_argument(*args, **kwargs)
        try:
            from argcomplete.completers import SuppressCompleter
        except ImportError:
            pass
        else:
            arg.completer = SuppressCompleter()
        return arg

    def main(self, *, context):  # noqa: D102
        args = context.args

        descriptors = get_package_descriptors(args)

        # always perform topological order for the select package extensions
        decorators = topological_order_packages(
            descriptors, recursive_categories=('run', ))

        select_package_decorators(args, decorators)

        if args.topological_graph or args.topological_graph_dot:
            additional_options = []
            if args.topological_graph_dot:
                if args.topological_graph_dot_cluster:
                    additional_options.append('--dot-cluster')
                if args.topological_graph_dot_include_skipped:
                    additional_options.append('--dot-include-skipped')
            elif args.topological_graph_density:
                additional_options.append('--density')
            if args.topological_graph_legend:
                additional_options.append('--legend')
            additional_options = ''.join(' ' + x for x in additional_options)
        if args.topological_graph:
            return 'The {args.verb_name} options --topological-graph / -g ' \
                'are obsolete, instead use `{context.command_name} graph' \
                '{additional_options}`'.format_map(locals())
        if args.topological_graph_dot:
            return 'The {args.verb_name} option --topological-graph-dot is ' \
                'obsolete, instead use `{context.command_name} graph --dot' \
                '{additional_options}`'.format_map(locals())
        if args.topological_graph_density:
            return 'The option --topological-graph-density must be used ' \
                'together with --topological-graph'
        if args.topological_graph_legend:
            return 'The option --topological-graph-legend must be used ' \
                ' with either --topological-graph or --topological-graph-dot'
        if args.topological_graph_dot_cluster:
            return 'The option --topological-graph-dot-cluster must be used ' \
                'together with --topological-graph-dot'
        if args.topological_graph_dot_include_skipped:
            return 'The option --topological-graph-dot-include-skipped must ' \
                'be used together with --topological-graph-dot'

        if not args.topological_order:
            decorators = sorted(
                decorators, key=lambda d: d.descriptor.name)
        lines = []
        for decorator in decorators:
            if not decorator.selected:
                continue
            pkg = decorator.descriptor
            if args.names_only:
                lines.append(pkg.name)
            elif args.paths_only:
                lines.append(str(pkg.path))
            else:
                lines.append(
                    pkg.name + '\t' + str(pkg.path) + '\t(%s)' % pkg.type)
        if not args.topological_order:
            # output names and / or paths in alphabetical order
            lines.sort()

        for line in lines:
            print(line)
